#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <strings.h>
#include <db.h>
#include "gensig.h"

extern int init_signals();

DB *open_db __P((char *, recno_t *));
int write_random_tagline __P((DB *, FILE *, recno_t));
int make_pidfile __P((void));
void detach __P((void));

int logout_flag;
int ppid;		/* Pid of parent (shell) */
char pidfile_name[BUFSIZ];

char *version = "gensig version 1.0, Apr. 19, 1995";

main()
{
	FILE *sig;		/* The .signature file */
	int err;
	DB *db;
	recno_t size;		/* # taglines in database */
	struct stat statb;
	time_t last_stat;

	logout_flag = 0;
	init_signals();		/* Catch various signals */
	ppid = getppid();	/* Get the shell's PID */

	detach();		/* Detach from the terminal and fork */

	srandom(time(NULL));	/* Seed the random number generator */

	if (make_pidfile() < 0)
		exit(1);

	/* Open the tagline database */
	if ((db = open_db(DB_FILE, &size)) == NULL)
	{
		unlink(pidfile_name);
		exit(1);
	}

	/* Stat the tagline database. See when it was last modified */
	stat(DB_FILE, &statb);
	last_stat = statb.st_mtime;


	alarm(TIMEOUT);		/* Set the alarm */

	/* Main loop. Basically, anything that should cause the
	 * process to exit will set 'logout_flag' to non-zero.
	 * This applies mainly to signal handlers, since the stuff
	 * inside the loop can just 'break'.
	 */
	while (!logout_flag)
	{
		if ((sig = fopen(SIGNATURE_FIFO, "w")) == NULL)
		{
			/* If it was interrupted by an alarm,
			 * ignore it and go back to bed.
			 */
			if ((errno == EINTR) && !logout_flag)
				continue;

			if (!logout_flag)
			{
				fprintf(stderr, "Can't open %s\n", SIGNATURE_FIFO);
				perror("fopen");
			}
			break;
		}

		/* Stat the tagline database to see if we need to
		 * reread it.
		 */
		stat(DB_FILE, &statb);
		if (statb.st_mtime != last_stat)
		{
			/* The file has been modified. Close and reopen
			 * it.
			 */
			db->close(db);
			if ((db = open_db(DB_FILE, &size)) == NULL)
			{
				break;
			}
			last_stat = statb.st_mtime;
		}

		fprintf(sig, "%s", PREAMBLE);
		write_random_tagline(db, sig, size);
		fprintf(sig, "%s", POSTAMBLE);
		fclose(sig);

		sleep(1);
	}

	/* Clean up and exit */
	db->close(db);
	fclose(sig);
	unlink(pidfile_name);
	exit(0);
}

/* open_db
 * Open the database 'dbname' for reading, and return a pointer to
 * the DB structure.
 * Sets 'nrecs' to the number of records in the database.
 */
DB *open_db(dbname, nrecs)
char *dbname;
recno_t *nrecs;
{
	DB *retval;
	DBT key, item;
	recno_t recnum;
	int err;

	/* Open the tagline database */
	if ((retval = dbopen(dbname, O_RDONLY, 0444, DB_RECNO, NULL)) == NULL)
	{
		fprintf(stderr, "open_db: Can't open database.\n");
		perror("dbopen");
		return(NULL);
	}

	/* Get the number of records in the database */
	/* Initialize 'key' and 'item' */
	key.data = &recnum;
	key.size = sizeof(recno_t);

	if ((err = retval->seq(retval, &key, &item, R_LAST)) < 0)
	{
		fprintf(stderr, "open_db: Can't seek to end.\n");
		perror("db->seq");
		retval->close(retval);
		return(NULL);
	}
	*nrecs = *((recno_t *) key.data);

	return(retval);
}

/* write_random_tagline
 * Pick a random tagline from the DB database 'db', and write it
 * to 'sig'.
 * 'nrecs' gives the number of records in the database.
 * All instances of \r in the tagline will be converted to \n before
 * printing.
 */
int write_random_tagline(db,sig,nrecords)
DB *db;
FILE *sig;
recno_t nrecords;
{
	DBT key, item;
	recno_t keynum;
	int err;
	char *ret;

	key.data = &keynum;
	key.size = sizeof(keynum);

	/* Pathological case: empty tagline database */
	if (nrecords <= 0)
		return(0);

	/* Pick a random tagline */
	keynum = (random() % nrecords)+1;

	if ((err = db->seq(db, &key, &item, R_CURSOR)) < 0)
	{
		fprintf(stderr, "write_random_tagline: Can't seek.\n");
		perror("db->seq");
		return(err);
	}
	/* Terminate the string, just to make sure */
	((char *) item.data)[item.size] = '\0';

	/* Replace \r with \n throughout */
	ret = item.data;
	while ((ret = index(ret, '\r')) != NULL)
		*ret++ = '\n';

	/* And finally print the tagline to 'sig' */
	fprintf(sig, "%s", (char *) item.data);

	return(0);
}

int make_pidfile()
{
	int myuid;		/* This user's UID */
	int mypid;		/* This process's PID */
	int pidfile_fd;		/* File descriptor for pidfile */
	static char buf[10];	/* 10 chars should be enough for a pid */

	myuid = getuid();
	mypid = getpid();
	sprintf(pidfile_name, "/tmp/.gensig.%d", myuid);

	/* Open the file. The flags are there in an attempt to
	 * make this an atomic lock.
	 */
	if ((pidfile_fd = open(pidfile_name, O_WRONLY|O_EXCL|O_CREAT, 0600))
		< 0)
	{
		if (errno == EEXIST)
		{
			/* Another gensig process is already running */
			printf("Another gensig process is already running.\n");
			return(-1);
		}

		/* Nope, it was a real error */
		perror("open");
		return(pidfile_fd);
	}

	/* Write this process's PID to the pidfile */
	sprintf(buf, "%d\n", mypid);
	if (write(pidfile_fd, buf, strlen(buf)) != strlen(buf))
	{
		unlink(pidfile_name);
		return(-1);
	}

	return(0);
}

/* detach()
 * Taken bodily out of zhm/zhm.c , in the zephyr sources.
 */
void detach()
{
	/* detach from terminal and fork. */
	register int i, size = getdtablesize();
  
	if (i = fork())
	{
		if (i < 0)
			perror("fork");
		exit(0);
	}

	for (i = 0; i < size; i++)
		(void) close(i);
  
	if ((i = open("/dev/tty", O_RDWR, 666)) < 0)
	;			/* Can't open tty, but don't flame about it. */
	else
	{
		(void) ioctl(i, TIOCNOTTY, (caddr_t) 0);
		(void) close(i);
	}
}
